-------------------------------------------------------------------------------------------
-- This include file keeps track of various actions to enable an UNDO/REDO function to work
-------------------------------------------------------------------------------------------

constant ACTIONLISTSIZE = 60

sequence actionlist, -- buffer to hold each event for all open files
        actionpointer -- points to last event actioned in list

--------------------------------------------------------------------------------------
global function isPreviousAction()
    if getNumberOfFiles()=0 then return w32False end if
    return actionpointer[getCurrentFileNum()]>0
end function

global function isNextAction()
    integer filenum
    if getNumberOfFiles()=0 then return w32False end if
    filenum = getCurrentFileNum()
    return actionpointer[filenum]<length(actionlist[filenum])
end function


global procedure updateUndoRedo()
    --# grey out/not undo & redo in the main menu
    if isPreviousAction() then
        setEnable(EditUndo, w32True)
    else
        setEnable(EditUndo, w32False)
    end if
    if isNextAction() then
        setEnable(EditRedo, w32True)
    else
        setEnable(EditRedo, w32False)
    end if
end procedure

global procedure resetActionList()
    actionlist = {}
    actionpointer = {}
    updateUndoRedo()
end procedure

resetActionList()


global procedure resetActions(integer filepos)
    actionlist[filepos] = {}
    actionpointer[filepos] = 0
    updateUndoRedo()
end procedure
--------------------------------------------------------------------------------------

global procedure addActionListForNewFile()
    actionlist = append(actionlist, {} )
    actionpointer &= 0
end procedure

global procedure deleteActionListForFile(integer file) -- delete actions list for that file
    integer len
    len = length(actionlist)

    if len = 1 then -- quick reset
        resetActionList()
        return
    end if

    actionlist[file .. len - 1] = actionlist[file + 1 .. len]
    actionlist = actionlist[1 .. len - 1]

    actionpointer[file .. len - 1] = actionpointer[file + 1 .. len]
    actionpointer = actionpointer[1 .. len - 1]

end procedure


function getLabellicLength(sequence date)
    integer len, c
    len = 0
    if length(date) = 1 then
        for i = 1 to length(date[1]) do
            c = date[1][i]
            if not isLabelCharacter(c) then
                return 0
            end if
            len += 1
        end for
    else
        return 0
    end if
    return len
end function


sequence mergeparams, mergedata


procedure mergeevents(integer msg, sequence params, sequence date)
    integer filenum, pointer

    integer msg_, len, len_
    sequence params_, data_

    mergeparams = params
    mergedata = date

    filenum = getCurrentFileNum()
    pointer = actionpointer[ filenum ]

    -- exit if no insert event to merge with
    if not pointer or msg = w32False then -- no prior events OR wrong action, ie delete
        return
    end if

    -- get values for current action
    msg_ = actionlist[filenum][pointer][1]
    params_ = actionlist[filenum][pointer][2]
    data_ = actionlist[filenum][pointer][3]

    -- determine length of 1-d strings
    len_ = getLabellicLength(data_)
    len = getLabellicLength(date)

    -- include terminating space with word
    if equal( {" "}, date ) then
        len = 1
    end if

    -- abort if situation not right
    if msg_ != msg -- messages different
    or not (len_ and len) -- old & new events do not both have linear label chars
    or params_[1] != params[1] -- old & new events NOT on sameline
        then return
    end if

    -- idiot test
    if params_[2] + len_ = params[2] then -- checksum ok
        mergedata = { data_[1] & date[1] } -- amalgamate data
        mergeparams = params_ + {0, 0, 0, len} -- expand params
        actionlist[filenum] = actionlist[filenum][1 .. length(actionlist[filenum]) - 1] -- remove end event now redundant
    end if

end procedure

global procedure addNewAction(integer msg, sequence params, sequence date) -- add event to list
    integer pointer, len, filenum

    filenum = getCurrentFileNum()
    pointer = actionpointer[ filenum ]

    -- does new action occur while a backtracking is happening?
    if pointer < length(actionlist[filenum]) then -- add event somewhere in middle
        actionlist[filenum] = actionlist[filenum][1 .. pointer] -- erase end events
    end if

    -- tack action to end of list (NEW!) or try to amalgamate with prev action
    mergeevents(msg, params, date)
    params = mergeparams 
    date = mergedata
    actionlist[filenum] = append(actionlist[filenum], {msg, params, date})

    -- ensure list does not exceed max size
    len = length(actionlist[filenum])
    if len > ACTIONLISTSIZE then
        actionlist[filenum] = actionlist[filenum][len - ACTIONLISTSIZE + 1 .. len]
    end if

    -- update pointer to point to the end
    actionpointer[filenum] = length(actionlist[filenum])
    updateUndoRedo()
end procedure

--------------------------------------------------------------------------------------
global function getNextAction()

    integer pointer, filenum

    filenum = getCurrentFileNum()
    pointer = actionpointer[ filenum ]

    if pointer < length(actionlist[filenum]) then -- at least one entry ahead
        pointer += 1
        actionpointer[filenum] = pointer
        updateUndoRedo()
        return actionlist[filenum][pointer]
    else
        return w32False
    end if

end function


--------------------------------------------------------------------------------------
global function getPreviousAction()

    integer pointer, filenum

    filenum = getCurrentFileNum()
    pointer = actionpointer[ filenum ]

    if pointer > 0 then -- at least one entry behind
        actionpointer[filenum] -= 1
        updateUndoRedo()
        return actionlist[ filenum ][pointer]
    else
        return w32False
    end if

end function

--------------------------------------------------------------------------------------
global procedure shuffleActs(integer direction)
--
-- swaps the action buffer info with that of file +/-1
--
    object temp
    integer c, cd
    c=getCurrentFileNum() -- hence should be before shuffleFile()
    cd=c+direction

    temp=actionlist[c] actionlist[c]=actionlist[cd] actionlist[cd]=temp
    temp=actionpointer[c] actionpointer[c]=actionpointer[cd] actionpointer[cd]=temp

end procedure