-------------------------------------------------------------------------------
---------------------------TEXT & COLOUR BUFFERS-------------------------------
-------------------------------------------------------------------------------

integer topline,    -- top line of text file shown on screen
        leftcolumn, -- leftmost column shown on screen
        tabwidth,
        maxrows     -- the maximum number of rows that can exist in the buffers

sequence textbuffer, -- holds the raw tab-corrected text data for the current screen
         colourbuffer -- holds the colour data for the current screen

-- give some initial value
textbuffer = {}
colourbuffer = {}

-- initialize tabwidth to force total initialization - is this needed?
    topline = 1
    leftcolumn = 1
    tabwidth = getTabWidth()

-------------------------------------------------------------------------------
   
global procedure setBufferLeftColumn(integer newleft)
    leftcolumn = newleft
end procedure

global procedure setBufferTopLine(integer newtop)
    topline = newtop
end procedure

global function getBufferLength()
    return length(textbuffer)
end function
-------------------------------------------------------------------------------


global procedure setBufferLines(integer top, integer bottom)
-- re-appoint the lines in the buffers

    sequence junk

    -- make sure overflow does not occur
    if bottom > length(textbuffer) then
        bottom = length(textbuffer)
    end if

    for i = top to bottom do

        -- prepare data
        junk = getAnalysedLine( i + topline - 1)
        junk = makeBufferLine(junk, tabwidth)

        -- insert into the buffers
        textbuffer[i] = junk[1]
        colourbuffer[i] = junk[2]

    end for

end procedure

-------------------------------------------------------------------------------

global procedure initializeBuffers(integer newtop, integer newleft, integer newrows, integer newtab)
-- totally refresh the data held in the buffers

    integer len

    -- store new values for object pointers
    topline = newtop
    leftcolumn = newleft
    tabwidth = newtab
    maxrows = newrows

    -- calculate max length of buffers
    len = getNumberOfLinesInFile() - topline + 1
    if len > maxrows then
        len = maxrows
    end if

    -- clear the buffer variables
    textbuffer = repeat(0, len)
    colourbuffer = repeat(0, len)

    -- appoint the lines in the buffers
    setBufferLines(1, len)
    
end procedure

-------------------------------------------------------------------------------

global procedure resizeBuffers(integer newrows, integer newtab)
    -- resize the buffers to the new length (new number of rows)

    integer len, screenlength

    -- prepare screen param and clip boundary pointer as needed
    screenlength = getNumberOfLinesInFile() - topline + 1
    if screenlength > newrows then
        screenlength = newrows
    end if

    if newtab != tabwidth then -- major reconstruction needed
        initializeBuffers(topline, leftcolumn, newrows, newtab)
        return
    end if
        
    len = length(textbuffer)

    -- any trimming required?
    if len > screenlength then
        textbuffer = textbuffer[1 .. screenlength]
        colourbuffer = colourbuffer[1 .. screenlength]

    -- OK, is extending required?
    elsif len < screenlength then
        textbuffer &= repeat(0, screenlength - len)
        colourbuffer &= repeat(0, screenlength - len)
        setBufferLines (len + 1, screenlength)

    end if

    maxrows = newrows
    tabwidth = newtab

end procedure

-------------------------------------------------------------------------------

global procedure scrollBuffers(integer newtop)
-- Refreshes only a change in the topline of the buffers.
-- It does not take into account a change in number of rows or tabs

    integer screenlength, bottom, newbottom, disp, len
    sequence junk

    -- do not process if no files
    if getNumberOfFiles() = 0 then
        return
    end if

    -- prepare screen param
    screenlength = getNumberOfLinesInFile() - newtop + 1
    if screenlength > maxrows then
        screenlength = maxrows
    end if

    -- buffers need to be shortened?
    if screenlength < length(textbuffer) then -- shorten buffers
        textbuffer = textbuffer[1 .. screenlength]
        colourbuffer = colourbuffer[1 .. screenlength]
    end if

    -- calc. new boundaries
    len = length(textbuffer)
    bottom = topline + len - 1
    newbottom = newtop + len - 1

    -- is new screen totally above old screen or totally under old screen?
    if newbottom < topline  or bottom < newtop then
        initializeBuffers(newtop, leftcolumn, maxrows, tabwidth)
        return
    end if

    -- SCROLLING

    -- does buffer need to scroll down?
    if topline < newtop then
        disp = newtop - topline
        -- move lowerhalf up
        textbuffer[1 .. len - disp] = textbuffer[disp + 1 .. len]
        colourbuffer[1 .. len - disp] = colourbuffer[disp + 1 .. len]
        -- fill in vacant space at bottom of buffers
        for i = len - disp + 1 to len do
            junk = getAnalysedLine( i + newtop - 1 )
            junk = makeBufferLine(junk, tabwidth)
            textbuffer[i] = junk[1]
            colourbuffer[i] = junk[2]
        end for

    -- ok, what about scrolling buffer up?
    elsif topline > newtop then
        disp = topline - newtop
        -- move upperhalf down
        textbuffer[disp + 1 .. len] = textbuffer[1 .. len - disp]
        colourbuffer[disp + 1 .. len] = colourbuffer[1 .. len - disp]
        -- fill in vacant space at top of buffers
        for i = 1 to disp do
            junk = getAnalysedLine( i + newtop - 1 )
            junk = makeBufferLine(junk, tabwidth)
            textbuffer[i] = junk[1]
            colourbuffer[i] = junk[2]
        end for

    end if

    -- add on extra lines?
    if len < screenlength then
        for i = len + 1 to screenlength do
            junk = getAnalysedLine(newtop + i - 1 )
            junk = makeBufferLine(junk, tabwidth)
            textbuffer = append(textbuffer, junk[1])
            colourbuffer = append(colourbuffer, junk[2])
        end for
    end if

    --- update line pointer
    topline = newtop

end procedure

-------------------------------------------------------------------------------

global procedure deleteBufferLines(integer cavitytop, integer cavitybottom)
-- chop a block of lines out of the buffers
    integer disp, len

    len = length(textbuffer)
    disp = cavitybottom - cavitytop + 1

    -- move lower half up
    textbuffer[cavitytop .. len - disp] = textbuffer[cavitybottom + 1 .. len]
    colourbuffer[cavitytop .. len - disp] = colourbuffer[cavitybottom + 1 .. len]

    -- truncate lower lines
    textbuffer = textbuffer[1 .. len - disp]
    colourbuffer = colourbuffer[1 .. len - disp]

    -- add new lines on to end as needed
    resizeBuffers(maxrows, tabwidth)

end procedure


global procedure insertBufferLines(integer cavitytop, integer cavitybottom)
-- insert a block of (space) lines into the buffers

    integer disp, len

    disp = cavitybottom - cavitytop + 1

    -- extend buffers
    textbuffer &= repeat(0, disp)
    colourbuffer &= repeat(0, disp)

    -- move lower half down
    len = length(textbuffer)
    textbuffer[cavitybottom + 1 .. len] = textbuffer[cavitytop .. len - disp]
    colourbuffer[cavitybottom + 1 .. len] = colourbuffer[cavitytop .. len - disp]

    -- trim lines as needed
    resizeBuffers(maxrows, tabwidth)

end procedure

-------------------------------------------------------------------------------

function makeBufferPrintLine(integer bLine, integer left, integer right)
-- extract the printable information from the buffer line in between the boundaries

    sequence textline, colourline, words, colours, backcolours, junk
    integer len, c, t, currentcolour

    textline = textbuffer[bLine]
    colourline = colourbuffer[bLine]
    len = length(textline)

    -- early out possible?
    if left > len then -- line not visible or nil-length
        return { {}, {}, {}}
    end if

    -- trim line ends
    if right > len then
        textline = textline[left .. len]
        colourline = colourline[left .. len]
    else
        textline = textline[left .. right]
        colourline = colourline[left .. right]
    end if

    -- initialize scratch vars
    junk = {}
    words = {}
    colours = {}
    backcolours = {}

    -- prepare start values
    junk &= textline[1]
    currentcolour = colourline[1]
    if currentcolour > COLOURTOP then
        currentcolour = COLOURHILITE
    end if

    if not syntaxON then
        if currentcolour>=COLOURKEYWORD and currentcolour<=COLOURTOP then
            currentcolour=COLOUROTHER
        end if
    end if

    -- now, deal to rest of data
    for i = 2 to length(textline) do

        t = textline[i]
        c = colourline[i]

        if c > COLOURTOP then
            c = COLOURHILITE
        end if
        if not syntaxON then
            if c>=COLOURKEYWORD and c<=COLOURTOP then
                c=COLOUROTHER
            end if
        end if
        if c != currentcolour then      -- change of colour
            -- store current assembled word
            words = append(words, junk)
            -- store colour to be associated with the word
            if currentcolour = COLOURHILITE then
                colours &= getColour(FOREGROUNDHILITE)
                backcolours &= getColour(BACKGROUNDHILITE)
            else
                colours &= getColour(currentcolour)
                backcolours &= getColour(COLOURBACKGROUND)
            end if
            -- set scratch vars with beginnings of next word/colour combo
            currentcolour = c
            junk = {t}
        else -- otherwise, just increase current word
            junk &= t
        end if

    end for

    -- tidy last pieces
    words = append(words, junk)
    if currentcolour = COLOURHILITE then
        colours &= getColour(FOREGROUNDHILITE)
        backcolours &= getColour(BACKGROUNDHILITE)
    else
        colours &= getColour(currentcolour)
        backcolours &= getColour(COLOURBACKGROUND)
    end if

    -- exit
    return {words, colours, backcolours}

end function

-------------------------------------------------------------------------------

global function makeBufferPrintBlock(integer top, integer left, integer bottom, integer right)
-- using the data refreshed in the buffers make an efficient print block for the pixmaps

    sequence result, junk

    left += leftcolumn - 1
    right += leftcolumn - 1
    result = {}

    for i = top to bottom do

        if i > length(textbuffer) then -- this may occur sometimes
            exit
        end if

        junk = makeBufferPrintLine(i, left, right)
        result = append(result, junk)

    end for

    return result
end function

-------------------------------------------------------------------------------

global procedure unhighlightBufferLines(integer top, integer bottom)
    for i = top to bottom do
        colourbuffer[i] -= (colourbuffer[i] > COLOURTOP) * COLOURTOP
    end for
end procedure

procedure highlightBufferLine(integer n, integer left, integer right)
-- highlight a single line in the colour buffer ensuring that no subscript error occurs
    if 0 < left and ((left <= right) or (left=1 and right=0)) and right <= length(colourbuffer[n]) then
        colourbuffer[n][left..right] += COLOURTOP
    end if
end procedure

global procedure highlightBufferLines( integer bTop, integer bBottom )
-- if the highlighting falls in between the bounds then highlight the relevant lines
    integer sTop, sLeft, sBottom, sRight

    -- early out
    if isTextSelected() = w32False then
        return
    end if

    -- get selectors
    sTop = getSelectTop()
    sLeft = getSelectLeft()
    sBottom = getSelectBottom()
    sRight = getSelectRight()

    -- clip lower bound
    if bBottom > length(colourbuffer) then
        bBottom = length(colourbuffer)
    end if

    -- make relative to screen
    sTop -= topline - 1
    sBottom -= topline - 1

    -- clip properly
    if sTop < bTop then
        sTop = bTop
        sLeft = 1
    end if
    if sBottom > bBottom then
        sBottom = bBottom
        sRight = length( colourbuffer[sBottom] )
    end if

    -- unhighlight colour buffer lines that falls in between the boundaries
    unhighlightBufferLines(bTop, bBottom)

    if sTop > sBottom then return end if

    -- processing single line?
    if sTop = sBottom then -- all on one line
        highlightBufferLine(sTop, sLeft, sRight)
    else
        -- highlight top line
        highlightBufferLine(sTop, sLeft, length(colourbuffer[sTop]) )

        -- highlight bottom line
        highlightBufferLine(sBottom, 1, sRight)

        -- highlight any lines in between
        for i = sTop + 1 to sBottom - 1 do
            colourbuffer[i] += COLOURTOP
        end for
    end if

end procedure

-------------------------------------------------------------------------------

global procedure highlightBuffer()
    integer top, bottom

    if isTextSelected() = w32False then
        return
    end if

    -- retrieve select params
    top = getSelectTop()
    bottom = getSelectBottom()

    -- make screen relative
    top -= topline - 1
    bottom -= topline - 1

    -- outside of screen?
    if top > length(colourbuffer) or bottom < 1 then
        return
    end if

    -- clip to screen size
    if top < 1 then
        top = 1
    end if
    if bottom > length(colourbuffer) then
        bottom = length(colourbuffer)
    end if

    highlightBufferLines( top, bottom )
end procedure

-------------------------------------------------------------------------------